"""General programming utils, inclined toward functional programming.

If ever a function changes its input in place, it is denoted by a trailing `_`
"""

import inspect
from itertools import zip_longest
from typing import List, Set, Union, Dict


def ifnone(*xs):
    """Return the first item in 'x' that is not None"""
    for x in xs: 
        if x is not None: return x
    return None

def custom_dir(c, add): return dir(type(c)) + list(c.__dict__.keys()) + add

class GetAttr:
    """Base class for attr accesses in `self._xtra` passed down to `self.default`
    
    Taken from article by Jeremy Howard: https://www.fast.ai/2019/08/06/delegation/

    Usage:

        ```
        class ProductPage(GetAttr):
            def __init__(self, page, price, cost):
                self.page,self.price,self.cost = page,price,cost
                self.default = page
        ```
    """
    @property
    def _xtra(self): return [o for o in dir(self.default) if not o.startswith('_')]
    def __getattr__(self,k):
        if k in self._xtra: return getattr(self.default, k)
        raise AttributeError(k)
    def __dir__(self): return custom_dir(self, self._xtra)

# Can i delegate many different functions?
# Can i add a new docstring to the existing docstring of the delgated function? Or at least point to the function delegated?
def delegates(to=None, keep=False):
    """ Decorator: replace `**kwargs` in signature with params from `to`.
    
    Taken from article by Jeremy Howard: https://www.fast.ai/2019/08/06/delegation/
    """
    
    def _f(f):
        if to is None: to_f,from_f = f.__base__.__init__,f.__init__
        else:          to_f,from_f = to,f
        sig = inspect.signature(from_f)
        sigd = dict(sig.parameters)
        k = sigd.pop('kwargs')
        s2 = {k:v for k,v in inspect.signature(to_f).parameters.items()
              if v.default != inspect.Parameter.empty and k not in sigd}
        sigd.update(s2)
        if keep: sigd['kwargs'] = k
        from_f.__signature__ = sig.replace(parameters=sigd.values())
        return f
    return _f

def pick(keys:Union[List, Set], obj:Dict) -> Dict:
    """ Return a NEW object containing `keys` from the original `obj` """
    return {k: obj[k] for k in keys}

def memoize(f):
    """Memoize a function.
    
    Use lookup table when the same inputs are passed to the function instead of running that function again
    """
    memo = {}
    def helper(*x):
        if x not in memo:            
            memo[x] = f(*x)
        return memo[x]
    return helper

def assoc(k, v, orig):
    """Given an original dictionary orig, return a cloned dictionary with `k` set to `v`"""
    out = orig.copy()
    out[k] = v
    return out

def make_unique(f):
    """The input function will only run and return if it hasn't seen its argument before. 
    
    Otherwise, it will return `None`.
    """
    s = set()
    def helper(x):
        if x in s:
            return None
        s.add(x)
        return f(x)

    return helper

def flatten_(items, seqtypes=(list, tuple)):
    """Flattten an arbitrarily nested list IN PLACE"""
    for i, x in enumerate(items):
        while i < len(items) and isinstance(items[i], seqtypes):
            items[i:i+1] = items[i]
    return items


"""
Describes the structure of a language token represented by Spacy-extracted metadata

"""
import h5py
import numpy as np
from spacy.tokens.token import Token as SpacyToken
from typing import Union, List, Tuple


def check_ent(tok: SpacyToken):
    """Check whether token is an entity

    Default Spacy Token does not assume what kind of entity you are looking for, but
    provides the following denotations:

        0: No entity tag is set
        1: inside an entity
        2: outside an entity
        3: Token begins an entity
    
    Args:
        tok: The Spacy Token

    Returns:
        Boolean indicating whether or not token is an entity    
    """
    OUT_OF_ENT = 2
    NO_ENT_DEFINED = 0
    return tok.ent_iob != OUT_OF_ENT and tok.ent_iob != NO_ENT_DEFINED

class SimpleSpacyToken():
    """A wrapper around a Spacy token to extract desired information

    This class implements a basic functional dictionary-like wrapper around the spacy token to 
    make it easy to mutate and export attributes without directly changing state. Any attribute
    that is not prefixed by '_' is considered a key of this class.

    The design allows for the token to have no metadata by simply passing a `str` into
    the constructor.

    Attributes:
        token: str
        pos: str
        dep: str
        norm: str
        tag: str
        lemma: str
        head: str
        is_ent: bool

    Notes:
        If exporting to an HDF5 file, make sure to define what hdf5 datatype that attribute 
        represents by changing the corresponding tuple in 'hdf5_token_dtype'
    """

    # Define how each attribute is stored in an hdf5 file 
    # Names MUST match attributes of this class
    hdf5_token_dtype = [
        ("token", h5py.special_dtype(vlen=str)),
        ("pos", h5py.special_dtype(vlen=str)),
        ("dep", h5py.special_dtype(vlen=str)),
        ("norm", h5py.special_dtype(vlen=str)),
        ("tag", h5py.special_dtype(vlen=str)),
        ("lemma", h5py.special_dtype(vlen=str)),
        ("head", h5py.special_dtype(vlen=str)),
        ("is_ent", np.bool_),
    ]

    def __init__(self, t:Union[SpacyToken, str]):
        """Create a simplified version of a spacy token
        
        Args:
            t: A string or Spacy Token object to wrap

        Raises:
            ValueError: If input is not of type SpacyToken or str
        """
        self._orig_token = t

        if type(t) == SpacyToken:
            self.token = t.text
            self.pos = t.pos_
            self.dep = t.dep_
            self.norm = t.norm_
            self.tag = t.tag_
            self.lemma = t.lemma_
            self.head = t.head
            self.is_ent = check_ent(t)

        elif type(t) == str:
            self.token = t
            self.pos = None
            self.dep = None
            self.norm = None
            self.tag = None
            self.lemma = None
            self.head = None
            self.is_ent = None

        else:
            raise ValueError("Expected input of SpacyToken or str")

    def pick(self, keys:List[str]):
        """Return subset of the attributes specified in 'keys' as a simple dictioniary
        
        Args:
            keys: List of keys to extract

        Returns:
            Dictionary of only k in keys

        Raises:
            KeyError: If k in 'keys' is not an attribute

        """
        return {k: self[k] for k in keys}

    def assoc(self, key:str, value):
        """Set the 'key' to the 'value', returning a new instance of this class.
        
        Args:
            key: Key that receives the value
            value: Value to assign to the key

        Returns:
            A new instance of this class with the modified key:value pair
        """
        out = SimpleSpacyToken(self._orig_token)
        out[key] = value
        return out

    def __getitem__(self, key):
        """Access the key from this objects dictionary"""
        return self.__dict__[key]

    def __setitem__(self, key, value):
        """Assign, in place, the value to the key"""
        self.__dict__[key] = value

    def keys(self) -> List[str]:
        """Return a list of all attributes that don't start with '_'"""
        return [k for k in self.__dict__.keys() if not k.startswith('_')]

    def values(self) -> List:
        """Return a list of all values whose keys don't start with '_'"""
        
        return [v for _, v in self.__dict__.items() if not k.startswith('_')]

    def items(self) -> List[Tuple]:
        """Return a list of all items whose keys don't start with '_'"""
        return [(k, v) for k,v in self.__dict__.items() if not k.startswith('_')]

    def __repr__(self):
        return f"SimpleSpacyToken: {self.items()}"


from .aligner import (
    MakeAligner,
    BertAligner,
    GPT2Aligner,
    RobertaAligner,
    DistilBertAligner,
    TransfoXLAligner,
    XLNetAligner,
    AlbertAligner,
    XLMAligner,
    CTRLAligner,
    OpenAIGPTAligner,
    T5Aligner,
    XLMRobertaAligner,
)

from .simple_spacy_token import SimpleSpacyToken
x
__all__ = ["MakeAligner", "SimpleSpacyToken", "BertAligner", "GPT2Aligner", "RobertaAligner", "DistilBertAligner",
           "TransfoXLAligner", "XLNetAligner", "AlbertAligner", "XLMAligner", "AlbertAligner",
           "CTRLAligner", "OpenAIGPTAligner", "T5Aligner", "XLMRobertaAligner"]



from .aligner import (
    BertAligner,
    GPT2Aligner,
    RobertaAligner,
    DistilBertAligner,
)

from .simple_spacy_token import SimpleSpacyToken

__all__ = ["SimpleSpacyToken", "BertAligner", "GPT2Aligner", "RobertaAligner", "DistilBertAligner"]



from setuptools import setup, find_packages

requires = [
    'transformers>=2.3.0',
    'h5py>=2.10.0',
    'numpy>=1.17.4',
    'regex>=2020.1.8',
]

setup(
    name="spacyface",
    description="Aligner for spacy and huggingface tokenization",
    packages=['spacyface'],
    version='0.1.1',
    license='MIT',
    author="Ben Hoover",
    author_email="benjamin.hoover@ibm.com",
    url="https://github.com/bhoov/spacyface",
    keywords=["transformer", "pytorch", "spacy", "tokenize", "tokenization", "NLP", "Natural Language Processing",
              "huggingface", "linguistic"],
    include_package_data=True,
    install_requires=requires,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        ]
)



from typing import List, Iterable, Union
import spacy
from spacy.tokens.token import Token as SpacyToken
from spacy.tokens.doc import Doc as SpacyDoc
import torch
import regex as re

from transformers import (
    BertTokenizer,
    GPT2Tokenizer,
    RobertaTokenizer,
    DistilBertTokenizer,
    TransfoXLTokenizer,
    XLNetTokenizer,
    XLMTokenizer,
    AlbertTokenizer,
    CTRLTokenizer,
    T5Tokenizer,
    XLMRobertaTokenizer,
    OpenAIGPTTokenizer,
    XLMRobertaTokenizer,
    AutoTokenizer,
)

from .simple_spacy_token import SimpleSpacyToken
from .utils.f import flatten_, assoc, delegates

def doc_to_fixed_tokens(doc: SpacyDoc) -> List[str]:
    """Fix the tokens in a document to not have exceptions"""
    return [fix_token(t) for t in doc]

def fix_token(tok: SpacyToken) -> str:
    """Determine whether a token should be represented by its text or its norm

    This works to fix most instances EXCEPT when an exception is part of a word with a '-' in it.
    For example, "whatve-you-done" would produce two different tokenizations:

    >>> alnr = BertAligner.from_pretrained('bert-base-uncased')
    >>> s = "whatve-you-dont"
    >>> alnr.tokenize(s) # => ['what', '##ve', '-', 'you', '-', 'don', '##t']
    >>> [t.token for t in alnr.meta_tokenize(s)] # => ['what', 'have', '-', 'you', '-', 'do', 'not']

    In practice, this situation occurs so rarely that it is often not a problem for real sentences to analyze.
    """
    out = tok.text if tok.text.lower() == tok.norm_ else tok.norm_

    return out

def MakeAligner(pretrained_tokenizer, spacy_language_model):
    """Create an aligner from the pretrained tokenizers. Some caveats to note:

    Usage:
        BrandNewHuggingfaceAligner = MakeAligner(BrandNewHuggingfaceTokenizer)
    """
    class Aligner(pretrained_tokenizer):
        @delegates(pretrained_tokenizer.__init__)
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.spacy_nlp = spacy.load(spacy_language_model)
            self.meta_container = SimpleSpacyToken

        def prep_sentence(self, s: str) -> str:
            """Remove contractions and multiple spaces from input sentence"""
            s = re.sub(r"\s+", r" ", s).strip()
            out = " ".join(self._to_normed_spacy(s))
            return out

        @delegates(pretrained_tokenizer.tokenize)
        def tokenize(self, s: str, **kwargs) -> List[str]:
            s = self.prep_sentence(s)
            return super().tokenize(s, **kwargs)

        def meta_tokenize(self, s: str) -> List[SimpleSpacyToken]:
            """Tokenize the sentence and return the metadata for it according to Spacy

            Due to implementation differences, does not provide the exact same API as the
            PreTrainedTokenizer's `tokenize` function
            """
            meta_info = self._to_spacy_meta(self.prep_sentence(s))
            return self._tokenize_from_spacy_meta(meta_info)

        def meta_from_tokens(self, sentence: str, tokens: List[str], perform_check=True) -> List[SimpleSpacyToken]:
            """Convert existing tokens into their metadata, ignoring effects of special tokens from the tokenizer

            NOTE that the sentence MUST be the same sentence that produced the tokens, otherwise,
            an unpredictable error may occur. Or worse, it will act like it works when it in fact doesn't.

            Parameters:
                - sentence: Sentence the tokens came from
                - tokens: Tokenized version of the sentence. Can be post encoding or pre-encoding
                    (where special tokens are added)
                - perform_check: If True, check that the tokens come from the sentence. This slows down processing
                    and should be False if speed is more important than accuracy
            """
            orig_meta = self.meta_tokenize(sentence)

            new_meta = []
            j = 0

            # Unfortunately, this can really slow down predictions.
            if perform_check:
                is_encoded = self.encode(sentence) == self.convert_tokens_to_ids(tokens)
                is_tokenized = self.tokenize(sentence) == tokens
                assert is_encoded or is_tokenized, "Can only take tokens that come from the original sentence!"

            for i, b in enumerate(tokens):
                if b in self.all_special_tokens:
                    new_meta.append(self.meta_container(b))
                else:
                    new_meta.append(orig_meta[j])
                    j += 1

            return new_meta

        def _to_normed_spacy(self, s: str) -> List[str]:
            """Return the normalized tokens (i.e., language exceptions replaced by a lowercased version)"""
            doc = self.spacy_nlp(s)
            tokens = self._doc_to_fixed_tokens(doc)
            return tokens

        def _to_spacy_meta(self, s: str) -> List[SimpleSpacyToken]: # list of simple spacy tokens...
            """Convert a string into a list of records containing simplified spacy information"""
            doc = self.spacy_nlp(s)
            out = [self.meta_container(t) for t in doc]
            return out

        @delegates(pretrained_tokenizer.tokenize)
        def _raw_tokenize(self, s: str, **kwargs) -> List[str]:
            """This bypasses the custom tokenization for the tokenization of the original model."""
            return super().tokenize(s, **kwargs)

        def _to_raw_spacy(self, s: str) -> List[str]:
            """Return the raw spacy tokens of a string"""
            doc = self.spacy_nlp(s)
            tokens = [t.text for t in doc]
            return tokens

        def _tokenize_from_spacy_meta(self, spacy_meta: List[SimpleSpacyToken]) -> List[SimpleSpacyToken]:
            """Convert spacy-tokenized SimpleSpacyTokens into the appropriate tokenizer's tokens"""
            out = [self._tokenize_from_meta_single(sm, i) for i, sm in enumerate(spacy_meta)]
            return flatten_(out)

        def _tokenize_from_meta_single(self, meta_token: SimpleSpacyToken, idx:int) -> List[SimpleSpacyToken]:
            """Split a single spacy token with metadata into tokenizer tokens.

            Because the transformer's tokenizer may split each Spacy-tokenized word into multiple subwords,
            output a list

            For GPT2 tokenization, there is a different behavior for the tokenization of a word if it
            starts the sentence vs if it occurs later in the sentence.
            """
            BUFFER = "X " # GPT tokenization fusses if it thinks the token is the beginning of the sentence

            def choose_norm(t):
                return t['token'] if t['token'].lower() == t['norm'] else t['norm']

            tok = choose_norm(meta_token)

            if idx != 0:
                s = BUFFER + tok # Add a buffer with guaranteed tokenization of length 1 to input
                offset = 1
            else:
                s = tok
                offset = 0

            bpe_tokens = super().tokenize(s) # Can't do `self.tokenize` because it will normalize again

            # Functional version that works with dictionaries
            return [meta_token.assoc("token", b) for b in bpe_tokens[offset:]]

        def _doc_to_fixed_tokens(self, doc: SpacyDoc) -> List[str]:
            """Extract tokens from a document, accounting for exceptions only if needed"""
            tokens = doc_to_fixed_tokens(doc)
            return tokens

        def _maybe_conv_to_token(self, tok_or_str:Union[str, SimpleSpacyToken]):
            """Convert a token to a SimpleSpacy token if a string. Otherwise, return input unmodified

            Args:
                tok_or_str: The token be analyzed

            Returns:
                SimpleSpacyToken. If input was a string, it has been converted to this class.
            """

            if isinstance(tok_or_str, SimpleSpacyToken):
                return tok_or_str
            return SimpleSpacyToken(self.convert_ids_to_tokens([tok_or_str])[0])

        def sentence_to_input(self, sentence:str):
            """Convert sentence to the input needed for a huggingface model

            Args:
                sentence: Sentence to prepare to send into the model

            Returns:
                Tuple of (object that can be directly passed into the model, modified meta tokens)

            Examples:

                >>> alnr = RobertaAligner.from_pretrained('roberta-base')
                >>> model = AutoModel.from_pretrained('roberta-base', output_attentions=True)
                >>> model.eval() # Remove DropOut effect
                >>> model_input, meta_info = alnr.sentence_to_input(sentence)
                >>> last_layer_hidden_state, pooler, atts = model(**model_input)
            """

            meta_tokens = self.meta_tokenize(sentence)
            tokens = [tok.token for tok in meta_tokens]
            ids = self.convert_tokens_to_ids(tokens)
            raw_model_input = self.prepare_for_model(ids, add_special_tokens=True)
            model_input = {k: torch.tensor(v).unsqueeze(0) for k,v in raw_model_input.items() if isinstance(v, List)}

            meta_input = self.prepare_for_model(meta_tokens)['input_ids']
            new_meta = list(map(self._maybe_conv_to_token, meta_input))

            return model_input, new_meta

        def check_tokenization(self, sentence:str, hard_assert=True):
            tokens = self.tokenize(sentence)
            meta_tokens = self.meta_tokenize(sentence)
            mtokens = [m.token for m in meta_tokens]

            error_str = """Meta tokenization did not match expected tokenization!

            EXPECTED:
            {}

            META TOKENS REPORTED:
            {}

            """
            is_fine = mtokens == tokens

            if hard_assert:
                assert is_fine, error_str.format(tokens, mtokens)
            else:
                if not is_fine: print(error_str.format(tokens, mtokens))

    return Aligner

english = "en_core_web_sm"

BertAligner = MakeAligner(BertTokenizer, english)
GPT2Aligner = MakeAligner(GPT2Tokenizer, english)
RobertaAligner = MakeAligner(RobertaTokenizer, english)
DistilBertAligner = MakeAligner(DistilBertTokenizer, english)
TransfoXLAligner = MakeAligner(TransfoXLTokenizer, english)
XLNetAligner = MakeAligner(XLNetTokenizer, english)
XLMAligner = MakeAligner(XLMTokenizer, english)
CTRLAligner = MakeAligner(CTRLTokenizer, english)
AlbertAligner = MakeAligner(AlbertTokenizer, english)
OpenAIGPTAligner= MakeAligner(OpenAIGPTTokenizer, english)
T5Aligner= MakeAligner(T5Tokenizer, english)
XLMRobertaAligner= MakeAligner(XLMRobertaTokenizer, english)
